#pragma once

#include <cstdint>

#ifdef POK_SIM
#include <SDL2/SDL.h>
extern SDL_AudioDeviceID audioDevice;
extern "C" void TIMER32_0_IRQHandler();
#else
#include "hw/Core.h"
#endif

namespace Pomi {

class Mixer {
    
    public:
    
        using Callback = std::int32_t (*)(void*);
        
        static constexpr std::uint32_t MATCHNUM = 1;
        
        Mixer() = delete;
        Mixer(const Mixer&) = delete;
        Mixer& operator =(const Mixer&) = delete;
        
        static inline void init(std::uint32_t samplerate) {
            #ifdef POK_SIM
            
            SDL_AudioSpec wanted;
            SDL_memset(&wanted, 0, sizeof(wanted));
            wanted.freq = samplerate;
            wanted.format = AUDIO_U8;
            wanted.channels = 1;
            wanted.samples = 512;
            wanted.callback = +[](void*, std::uint8_t* stream, int len){
                for(int i=0; i<len; ++i) {
                    TIMER32_0_IRQHandler();  // Process one audio sample and store result in variable _to_dac
                    stream[i] = _to_dac;
                }
            };

            audioDevice = SDL_OpenAudioDevice(NULL, 0, &wanted, NULL, 0);
            SDL_PauseAudioDevice(audioDevice, 0);
            
            #else
            
            Pomi::Core::enableDAC();
            Pomi::Core::enableTimer(0, MATCHNUM, samplerate);
            
            #endif
        }
        
        static inline void setChannel(std::uint8_t channel, Callback callback, void* data=nullptr) {
            // Temporarily change callback to nullSource before setting data, to protect against random crashes,
            // when audio IRQ executes just after setting data but before the new callback is set
            Channel& ch = _channels[channel];
            ch.callback = nullSource;
            ch.data = data;
            if(callback) ch.callback = callback;
        }
        
        // Sets the master volume. 256 equals unit gain.
        static inline void setVolume(std::uint32_t volume) {
            _gain_Q8 = (volume * volume) >> 8;
        }
        
        static inline void onTimerMatch() {
            #ifndef POK_SIM
            // Before anything else, the previous output value is written to the DAC.
            // This way DAC is fed at constant rate even if the processing time varies.
            Pomi::Core::writeDAC(_to_dac);
            #endif
            
            // Process all channels and mix them together
            std::int32_t out_Q15 = 0;
            for(std::uint32_t idx = 0; idx < 4; ++idx) {
                Channel& ch = _channels[idx];
                out_Q15 += ch.callback(ch.data);
            }
            
            // Convert output to unsigned 8 bit value
            _to_dac = 128 + (out_Q15 * _gain_Q8) / (1 << 16);
            _to_dac = (_to_dac > 255) ? 255 : ((_to_dac < 0) ? 0 : _to_dac);
        }
    
    private:
    
        struct Channel {
            constexpr Channel(): callback(nullSource), data(nullptr) {}
            
            Callback callback;
            void* data;
        };
        
        static constexpr std::int32_t nullSource(void* data) {
            return 0;
        }
        
        static std::int32_t _to_dac;
        static std::int32_t _gain_Q8;
        static Channel _channels[4];
};

} // namespace Pomi
